//(root)/Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7client.h - Rev 13706
//
// Gen7 Audio Client Library
// Copyright (c) 2010-2015 Mimosa Acoustics, Inc.

// Linux end-of-line markers, 8-column _tabs_ (not spaces), K&R C brace style.

// SOURCE CONTROL WARNING:  The master version of this file is located in the Gen7 repository at
//      Gen7/HandheldSourceCode/Overlay/recipes/g7common/g7client.h
// Subversion updates under Windows may result in duplicate copies due to following symbolic links,
// but all changes should be made and committed to the master version at the location showmn above.

#pragma once

// Need shared definitions used by client and server.
#include "g7shared.h"

// General approach:
//
// Power up:
// (1) Call g7_init() to create a client instance and set the debug mask and debug log path.
// (2) Call g7_control(G7_FLAG_CTRL_SAMPLES_PER_BLOCK) to specify the number of samples per block.
// (3) Call g7_connect() to connect to the server.

// Normal operation:
// (4) Otostat-C3: Call g7_control(G7_FLAG_CTRL_SETUP_C3) to set up the audio path.
//
// (5) Call g7_control(G7_FLAG_CTRL_RATE) to set the desired sample rate.
// (6) Call g7_modify_output() for each output channel to supply audio data.
// (7) Call g7_control_output() to begin audio output and capture.
// (8) Call g7_read_input() repeatedly until -1 is returned.
//
// (9) Otostat-C3: If desired, call g7_control(G7_FLAG_CTRL_SETUP_C3) to shut down the audio path
//     or put portions of the audio path in low-power mode.
//
// Power down:
// (10) Otostat-C3: Call g7_control(G7_FLAG_CTRL_SETUP_C3) to power down the audio path.
// (11) Call g7_connect() to disconnect.
// (12) Call g7_shutdown() to uninitialize client.

// Future extensions not yet supported:
//
// Means of identifying hardware and software version information from the peripheral.
// Means of getting the minimum and maximum block size from the peripheral.
// Means of getting the input and output channel counts from the peripheral.

// A handle to a specific g7client instance.
typedef void* g7handle_t;

#if defined(_MSC_VER)
#define G7CLIENT_EXPORT __declspec(dllexport)
#else
#define G7CLIENT_EXPORT
#endif

#if defined(__cplusplus)
extern "C" {
#endif

// Unless otherwise specified, functions return G7_OK on success, or a G7_ error code.
// See g7_errtext() for obtaining more detailed error information.

// Initialize the client.  This must be called prior to making any other library calls on this instance.
// mask_ is any combination of the G7CLIENT_DEBUG_ bit constants.
// log_path_, if non-NULL, is the path to the file to which messages should be written.
// If log_path_ is NULL or points to an empty string, output will be logged to stdout.
//
// If the call succeeds, it returns a handle suitable for passing to other g7_ functions.
// If the calls fails, it returns NULL.
G7CLIENT_EXPORT g7handle_t g7_init(unsigned int mask_, const char* log_path_);

// Get a pointer to a text string describing the last error which occurred.
// The returned pointer is valid until the next call to a g7_ function.
G7CLIENT_EXPORT const char* g7_errtext(g7handle_t handle_);

// Get a pointer to a text string describing G7_OK or a G7_ERR_ value.
// The returned pointer is valid until the next call to g7_returntext.
G7CLIENT_EXPORT const char* g7_returntext(g7handle_t handle_, int code);

// Get a description of the specified loopback code; returns default_ if unknown.
// handle_ is not currently used, and may be NULL.
G7CLIENT_EXPORT const char* g7_loopbacktext(g7handle_t handle_, int code_, const char* default_);

// Write text into the buffer returned by g7_errtext, using printf semantics.
// This is available for helper functions which want to "look like" g7_ functions.
// Such a function should call this function and then return a valid G7_ERR_ value.
// To clear the error text buffer, pass NULL in format_.
G7CLIENT_EXPORT void g7_set_errtext(g7handle_t handle__, const char* format_, ...);

// Set various control parameters.
//
// flags_ determines the meaning of the argument.  It may be one of the following:
//
//      Settable at any time after initialization:
//
//      G7_FLAG_CTRL_CONNECT_TIMEOUT    Set the connect timeout in milliseconds.  This is the time within which the
//                                      server must accept a connect() request from the client.  This can be set to
//                                      0 for infinite timeout (for debugging purposes).
//
//      G7_FLAG_CTRL_PROGRESS_TIMEOUT   Set the progress timeout in milliseconds.  This is the time within which the
//                                      server must send data when we are waiting for it, or must accept data when
//                                      we are sending it.  Each received or transmitted burst of data resets the
//                                      timer.  This can be set to 0 for infinite timeouts (for debugging purposes).
//
//      G7_FLAG_CTRL_CLIENT_DEBUG_MASK  Set the bitmap of G7CLIENT_DEBUG_ flags which control what classes of debug
//                                      messages are printed by the client.
//
//      Settable only when not connected to a server:
//
//      G7_FLAG_CTRL_SAMPLES_PER_BLOCK  Set the number of samples per block.  This must be done prior to connecting.
//
//      Settable only while connected to a server:
//
//      G7_FLAG_CTRL_RATE               arg_ specifies the desired sample rate in Hertz.  An error will be returned
//      (Otostat-B Only)                if the specified rate is not supported by the audio hardware.  Changes to
//                                      the rate take effect when g7_control_output(G7_FLAG_CO_PLAY) is issued.
//
//      G7_FLAG_CTRL_SETUP_C3           Modify the audio path setup.  arg_ is defined as follows.  Note that for
//      (Otostat-C3 only)               simulated (dummy) hardware, only bits 3..0 are significant, and are used
//                                      to define the sample rate.
//
//                                      Bit(s)  Description
//                                      [31:27] Reserved (0)
//                                      [26]    5V probe power (0 = off, 1 = on)
//                                      [25]    Microphone bias power (0 = off, 1 = on)
//                                      [24]    Microphone preamplifier (0 = disabled, 1 = enabled)
//                                      [23:20] Primary input state (0 = disabled, 1 = enabled/low-gain, 2 = enabled/high-gain)
//                                      [19:16] Auxiliary input state (0 = disabled, 1 = aux input, 2 = monitor A, 3 = monitor B)
//                                      [15:12] Receiver A state (0 = disabled, 1 = enabled/attenuated, 2 = enabled/unattenuated)
//                                      [11:8]  Receiver B state (0 = disabled, 1 = enabled/attenuated, 2 = enabled/unattenuated)
//                                      [3:0]   Master state (0 = off, 1 = standby, 2 = 16KHz, 3 = 24 KHz, 4 = 32 KHz, 5 = 48 KHz).
//                                      NOTE:  If Master state is 0 or 1, all other bits are ignored.
//
//      G7_FLAG_CTRL_LOOPBACK           arg_ specifies the loopback type.  Supported loopback values are:
//                                              G7_LOOPBACK_SERVER (the server simulates the codec)
//                                              G7_LOOPBACK_I2S (I2S data reflects the full 32 bits of what is being sent to the codec)
//                                              G7_LOOPBACK_NONE (I2S data reflects the output of the ADC; the low 8 bits are zero)

// arg_ is a numeric setting based on flags_.
G7CLIENT_EXPORT int g7_control(g7handle_t handle_, unsigned int flags_, int arg_);

// Open a connection to the audio server.
//
// address_ may be any of the following:
//      The IP address and port of a g7server in dotted-decimal notation.  Example: "192.168.0.1:1234"
//      NULL or "", in which case any existing server connection is disconnected.
G7CLIENT_EXPORT int g7_connect(g7handle_t handle_, const char* address_);

// Query various server parameters.
//
// flags_ may be one of the following:
//
//      G7_FLAG_QUERY_CHANNELS          *result1_ will receive the number of playback (output) channels supported by the server.
//                                      *result2_ will receive the number of of capture (input) channels supported by the server.
//
//      G7_FLAG_QUERY_PLAYBACK          *result1_ will receive one of the G7_QUERY_PLAYBACK_ values.
//                                      *result2_ will receive the tag of the current stage in the playback sequence, if applicable (*)
//                                      *result3_ will receive the number of times the current stage has been repeated, if applicable (*).
//                                      (*) if state is not G7_QUERY_PLAYBACK_PLAYING, these will be set to 0.
//
//      G7_FLAG_QUERY_SETUP_C3          Otostat-C3 Only.
//                                      *result1_ will receive the audio server status, as described for G7_FLAG_CTRL_SETUP_C3.
//                                      *result2_ will receive a bit mask describing which bits in result1_ are significant.
//
// result1_, result2_, result3_ and result4_ point to locations to receive result values based on flags_.  Any or all may be NULL.
G7CLIENT_EXPORT int g7_query(g7handle_t handle_, unsigned int flags_, unsigned int* result1_, unsigned int* result2_, unsigned int* result3_, unsigned int* result4_);

// Sequence-based output control:
//
// Sequence-based output control defines the output stream as a sequence of stages, each consisting of a data block
// played a specified number of times.  A sequence may include stages with indefinite repeat counts, which will be
// played continuously until explicitly told to stop playing, or move on to the next stage.  The stream can be fully
// specified before playback, or can be truncated and/or appended to while playback is in progress; the server will
// ensure that transitions happen on block boundaries.  The start of the capture stream can be deferred to skip
// start-up blocks, and stopped after reaching a particular stage, to avoid capturing ramp-up and ramp-down blocks.
// After the end of the sequence is reached, playback halts and the codec is powered off.  To avoid this behavior
// (i.e., keep the codec outputting silence until new audio data can be appended to the sequence) end the sequence
// with an indefinitely repeating, non-captured stage of silence.
//
// The legacy sequence
//      g7_set_data(0, data0);
//      g7_set_data(1, data1);
//      g7_play(silent_blocks, data_blocks);
//
// is essentially equivalent to
//      unsigned int* data[2] = { cloned-data0, cloned-data1 };
//      where   cloned-data0 -> codec output channel A = Left
//              cloned-data1 -> codec output channel B = Right
//
//      g7_stage_output(G7_FLAG_SO_RESET | G7_FLAG_SO_APPEND_DATA | G7_FLAG_SO_NO_CAPTURE, 2, NULL, silent_blocks, 0);
//      g7_stage_output(G7_FLAG_SO_APPEND_RAMP_UP | G7_FLAG_SO_NO_CAPTURE, 2, data, 1, 0);
//      g7_stage_output(G7_FLAG_SO_APPEND_DATA, 2, data, data_blocks, 0);
//      g7_stage_output(G7_FLAG_SO_APPEND_RAMP_DOWN | G7_FLAG_SO_NO_CAPTURE, 2, data, 1, 0);
//      g7_control_output(G7_FLAG_CO_START);
//
// The legacy sequence
//      g7_stop();
//
// is equivalent to
//      g7_control_output(G7_FLAG_CO_STOP | G7_FLAG_CO_DISCARD);
//      g7_stage_output(G7_FLAG_SO_RESET, 0, NULL, 0, 0);

// Create or modify a sequence of output stages for subsequent playback.
//
// flags_ is any combination of the following ORed together:
//
//      Zero or one of the following:
//
//      G7_FLAG_SO_APPEND_DATA                  The block specified by data_ is added to the end of the sequence as a new
//                                              stage, with the repeat count specified by repeat_.  data_ may be NULL which
//                                              is equivalent to specifying a block of all zeroes.
//
//      G7_FLAG_SO_APPEND_RAMP_UP               A linear ascending ramp is applied to data_ and appended as the new stage.
//                                              A repeat count of 1 is recommended.
//
//      G7_FLAG_SO_APPEND_RAMP_DOWN             A linear descending ramp is applied to data_ and appended as the new stage.
//                                              A repeat count of 1 is recommended.
//
//      Zero or more of the following:
//
//
//      G7_FLAG_SO_RESET                        Prior to taking any other action the staged output sequence is cleared.
//                                              This may be specified alone or in conjunction with any of the APPEND_ flags.
//
//      G7_FLAG_SO_NO_CAPTURE                   Flags the block as a "non-capture" block, as described in the following.
//                                              This is only meaningful in conjunction with one of the APPEND_ flags.
//                                              Blocks of capture data which are captured contemporaneously with blocks
//                                              thus flagged are discarded.  This allows initial silence, ramp-up blocks,
//                                              ramp-down blocks and transitions to be marked as non-capturable.
//
//      G7_FLAG_SO_START_OF_LOOP                Flags the block as the start of a loop.  If any samples have been output
//                                              from a block marked with this flag, when the end of the output sequence
//                                              is reached, instead of outputting silence, playback will continue with
//                                              the most-recently-output block marked with this flag.
//
// The following are significant only when one or more of the G7_SO_APPEND_ flags is specified:
//
// channel_ is the number of playback channels for which data is supplied.  Generally this should be the number of
// channels supported by the output hardware, although if fewer channels are supplied, silence is assumed for the
// remaining channels (equivalent to passing a NULL pointer in the remaining elements of the data_ pointer array).
//
// data_ consists of channel_ consecutive pointers, each pointing to a block of samples to be played.  The size of
// each block must match the samples-per-block count specified in the call to g7_set_block_size.  Pointers can be
// NULL in which case a block of zeroes is assumed.  Multiple pointers in data_ can point to the same source data.
//
// repeat_ is the number of times the specified block is to be output in order to complete this stage of the sequence.
// A value of 0 specifies an indefinite-length stage, which will be repeated indefinitely until playback is stopped,
// or until explicitly skipped by a g7_control_output call with the G7_FLAG_CO_CONTINUE flag specified.
//
// tag_ is a caller-provided value which is associated with the playback stage, and will be used to set the *tag_
// parameter from the g7_read_extended call, in order to indicate which playback sequence was captured.
G7CLIENT_EXPORT int g7_stage_output(g7handle_t handle_, unsigned int flags_, unsigned int channels_, const sample_t** data_, unsigned int repeat_, unsigned int tag_);

// Control the output stream.
//
// flags_ is any legal combination of the following ORed together (see descriptions):
//
//      G7_FLAG_CO_START                        Synchronize the audio streams and begin playing the staged sequence (if any).
//                                              Control returns after after synchronization has been achieved (or has failed).
//                                              No other flags may be specified with this option.
//                                             
//                                              On Otostat-B, if the audio hardware is not powered up and configured
//                                              for the sample rate requested with g7_control(G7_FLAG_CTRL_RATE), this call
//                                              will cause it to be powered up and configured for the requested rate.
//
//                                              On Otostat-C3, an error will be returned if the audio hardware has not been
//                                              set up with g7_control(G7_FLAG_CTRL_SETUP_C3) for a master state of 2, 3, 4, or 5.
//
//      G7_FLAG_CO_START_UNSYNCHRONIZED         Like G7_FLAG_CO_START, except audio synchronization is not performed.
//                                              As a result, capture blocks will not be aligned with playback blocks.
//                                              This is provided for testing, or to eliminate the delay associated with
//                                              synchronization for applications which do not need to associate capture
//                                              and playback data (e.g., playback-only or capture-only applications).
//                                              When running unsynchronized, all capture blocks will be reported with a
//                                              tag of all-ones and a sequence number incrementing up from 0.  Because
//                                              playback and capture blocks are unassociated, capture will continue
//                                              after the playback sequence is ended.  If the caller needs to stop the
//                                              transfer after the playback sequence has completed, the caller must
//                                              monitor the state of playback by by calling g7_query specifying the
//                                              G7_FLAG_QUERY_PLAYBACK flag, to determine when playback has completed.
//
//      G7_FLAG_CO_STOP                         Stop playback and power down the codec.  Control resumes after the codec
//                                              has been powered down.  Only the G7_FLAG_CO_DISCARD flag may be specified
//                                              in combination with this option.
//
//      G7_FLAG_CO_CONTINUE                     Advance to the next stage of playback on the next block boundary.  This is
//                                              used to skip a long or infinitely-repeating output stage.  Control returns
//                                              after at least one sample of the new stage has been output.  This flag can
//                                              be specified alone or in conjunction with _SPLICE, _TRUNCATE and/or _DISCARD.
//
//      NOTE: G7_FLAG_CO_CONTINUE will yield predictable results only when used during the output of a playback stage
//      of indefinite length.  Otherwise, one or more stages may end on their own due to expiration of their repeat
//      count before the continue operation takes effect, and a different stage than the one intended may be skipped.
//
//      G7_FLAG_CO_SPLICE                       The staged sequence is added to the end of the output sequence to be played.
//                                              If the output sequence is empty or at end-of-stream, the new sequence will
//                                              begin playing on the next block boundary; otherwise it will begin playing
//                                              when all stages prior to it have been completed and/or skipped.  Control
//                                              returns as soon as the output sequence has been modified.  This flag can be
//                                              specified alone or in conjunction with _CONTINUE, _TRUNCATE and/or _DISCARD.
//
//      G7_FLAG_CO_TRUNCATE                     All stages in the output sequence (beyond the stage currently being played)
//                                              are discarded.  If specified in conjunction with G7_FLAG_CO_SPLICE, it is
//                                              guaranteed that the truncation and splicing will happen together; that is,
//                                              any stages following the current stage will be discarded and replaced with
//                                              the staged sequence.  This flag can be specified alone or in conjunction
//                                              with _CONTINUE, _SPLICE and/or _DISCARD.
//
//      G7_FLAG_CO_DISCARD                      Discard any captured data.  This can be supplied on its own, or used in
//                                              combination any other request, although it is redundant in the case of
//                                              G7_FLAG_CO_START since capture data is always discarded prior to start).
//                                              Note that unlike g7_read_extended with the G7_FLAG_READ_DISCARD flag,
//                                              data discarded by this command does not honor tag boundaries; all data
//                                              will be discarded regardless of its tags.
//
//      NOTE:  If multiple of _CONTINUE, _SPLICE, _TRUNCATE, and/or _DISCARD are specified, the order of operations
//      is _DISCARD, _TRUNCATE, _SPLICE and _CONTINUE; that is, splicing is done onto the truncated output stream,
//      and _CONTINUE skips the stage currently being played and advances to the newly-spliced data (or the silence
//      at end of stream if no new data is available).
G7CLIENT_EXPORT int g7_control_output(g7handle_t handle_, unsigned int flags_);

// Multi-channel replacement for g7_read that provides additional information on the blocks being returned, and which does
// not read across tag boundaries (i.e., all blocks returned will have the same tag value).
//
// channels_ is a bitmap indicating the channels for which data is to be returned (MSB = channel 0).
//
// buffer_ is the caller-provided buffer to receive the samples, which must be non-NULL.  buffer_ must contain
// enough to space to hold one buffer for each bit set in channels_, times (*_blocks), times the block size set
// by g7_set_block_size, times sizeof(sample_t) (the size of each sample in the block).
//
// Blocks are returned in the order (ch0-blk0, ch1-blk0, .., chN-blk0, ch0-blk1, ch1-blk1, ...); that is, in
// time-major rather than channel-major order.  In future revisions a flag may be added to support deinterleaving.
//
// blocks_ points an in/out parameter.  The caller must set it to the maximum number of BLOCKS (not bytes) to be returned
// PER CHANNEL prior to each call.  (e.g., if there are 2 channels and the caller wants 3 blocks per channel, *blocks_
// should be set to 2, not 6).  Upon return (even an error return, when it will be set to zero), it will be set to the
// number of BLOCKS (not bytes) stored at buffer_ PER CHANNEL.
//
// flags_ is a combination of the following ORed together:
//
//      Zero or one of the following:
//
//      G7_FLAG_READ_LATEST                     Rather than the oldest available (*blocks_) blocks, return the most recent
//                                              available (*blocks_) blocks.  If N blocks are available and N <= (*blocks_),
//                                              all N blocks are returned.  If N blocks are available and N > (*blocks_),
//                                              the oldest N - (*blocks_) are discarded and the remaining N are be returned.
//
//      G7_FLAG_READ_DISCARD                    Discard up to the oldest (*blocks_) blocks.  In this case buffer_ must be NULL.
//                                              (*blocks_), and (*number_) are set as if the blocks were actually read, but
//                                              the blocks are not actually transferred; this is an efficient way to discard
//                                              blocks without to actually transfer them over the network.  Data will only be
//                                              discarded up to the next tag boundary (just as if the data were being read).
//                                              To discard all captured data without regard for tags, use g7_control_output
//                                              with the G7_FLAG_CO_DISCARD flag.
//
//      Zero or one of the following:
//
//      G7_FLAG_READ_24_RJ_IN_32                Return capture data as 24-bit samples right-justified and sign extended
//                                              in 32-bit words; that is, the highest order 8 bits are always 0x00 or 0xFF.
//                                              The default is 24-bit samples left-justified with the low-order 8 bits of 0.
//                                              Note that data is always passed to the codec left-justified; this controls
//                                              only how the data is transferred into the network buffer for transport.
//                                              This flag is obviously not meaningful with the G7_FLAG_READ_DISCARD flag.
//
// If tag_ is non_NULL, (*tag_) receives the value of the tag of the output sequence which was being output when this
// block was captured.  Since playback and capture are synchronized, there will thus be a direct correspondence
// between blocks from a given stage of a playback sequence, and the blocks captured while that stage was playing.
//
// If sequence_ is non-NULL, (*sequence_) receives the number of times the block in the output sequence whose tag is
// reflected by tag_ had been output at the time the first capture block was captured; that is, the zero-based index
// of this capture block within the stage of the output sequence which was playing when it was captured.
//
// NOTE: If multiple blocks are returned, they are always contiguous, and will never cross a tag boundary.
//
// NOTE: Attempting to read past end of stream will result in G7_ERR_END_OF_STREAM.  It is legitimate for the caller
// to keep calling g7_read_extended until G7_ERR_END_OF_STREAM is returned as an indication that data has been exhausted.
G7CLIENT_EXPORT int g7_read_input(g7handle_t handle_, unsigned int channels_, sample_t* buffer_, unsigned int* blocks_, unsigned int flags_, unsigned int* tag_, unsigned int* sequence_);

// Write a command sequence to the control processor.
//
// NOTE:  This is on the verge of obsolescence; the only thing it is used for is working with the real-time clock
// in the Rev B Otostat.  There is probably a better place for this functionaliy than g7server, but here it is.
//
// command_ points to the buffer containing the sequence of bytes to send to the control processor.
// byte_count_ specifies the number of bytes in the command sequence (which is also the number of bytes in the response).
// On a successful return, the byte_count_ response is stored at response_.
G7CLIENT_EXPORT int g7_link(g7handle_t handle_, unsigned char const* command_, unsigned int byte_count_, unsigned char* response_);

// Shut down the client.  If the client thinks that playback is in progress, it will try to stop it,
// but will continue with the shutdown even if the peripheral fails to respond or responds with an error.
// There is no error return; the client will do the best it can to shut down regardless of errors.
G7CLIENT_EXPORT void g7_shutdown(g7handle_t handle_);

#if defined(__cplusplus)
};
#endif
